from __future__ import annotations

__all__ = ("asimov_sig", "gaussianity")

from functools import partial
from typing import TYPE_CHECKING, cast

import jax
import jax.numpy as jnp
import jax.scipy as jsp
from jax.random import PRNGKey, multivariate_normal

from relaxed._types import Array
from relaxed.ops import fisher_info

if TYPE_CHECKING:
    import pyhf


@jax.jit
def asimov_sig(s: Array, b: Array) -> float:
    """Median expected significance for a counting experiment, valid in the asymptotic regime.
    Also valid for the multi-bin case.

    Parameters
    ----------
    s : Array
        Signal counts.
    b : Array
        Background counts.

    Returns
    -------
    float
        The expected significance.
    """
    q0 = 2 * jnp.sum((s + b) * (jnp.log(1 + s / b)) - s)
    return cast(float, q0**0.5)


def _gaussian_logpdf(
    bestfit_pars: Array,
    data: Array,
    cov: Array,
) -> Array:
    return jsp.stats.multivariate_normal.logpdf(data, bestfit_pars, cov).reshape(
        1,
    )


@partial(jax.jit, static_argnames=["model", "n_samples"])
def gaussianity(
    model: pyhf.Model,
    bestfit_pars: Array,
    data: Array,
    rng_key: PRNGKey,
    n_samples: int = 1000,
) -> Array:
    # - compare the likelihood of the fitted model with a gaussian approximation
    # that has the same MLE (fitted_pars)
    # - do this across a number of points in parspace (sampled from the gaussian approx)
    # and take the mean squared diff
    # - centre the values wrt the best-fit vals to scale the differences

    cov_approx = jnp.linalg.inv(fisher_info(model, bestfit_pars, data))

    gaussian_parspace_samples = multivariate_normal(
        key=rng_key,
        mean=bestfit_pars,
        cov=cov_approx,
        shape=(n_samples,),
    )

    relative_nlls_model = jax.vmap(
        lambda pars, data: -(
            model.logpdf(pars, data)[0] - model.logpdf(bestfit_pars, data)[0]
        ),  # scale origin to bestfit pars
        in_axes=(0, None),
    )(gaussian_parspace_samples, data)

    relative_nlls_gaussian = jax.vmap(
        lambda pars, data: -(
            _gaussian_logpdf(pars, data, cov_approx)[0]
            - _gaussian_logpdf(bestfit_pars, data, cov_approx)[0]
        ),  # data fixes the lhood shape
        in_axes=(0, None),
    )(gaussian_parspace_samples, bestfit_pars)

    diffs = relative_nlls_model - relative_nlls_gaussian
    return jnp.nanmean(diffs**2, axis=0)



from __future__ import annotations

__all__ = ("fit", "fixed_poi_fit")

from functools import partial
from typing import TYPE_CHECKING, Any, Callable, cast

import jax
import jax.numpy as jnp
import jaxopt
import optax

from relaxed._types import Array

if TYPE_CHECKING:
    import pyhf


@partial(jax.jit, static_argnames=["objective_fn"])
def _minimize(
    objective_fn: Callable[..., float], init_pars: Array, lr: float, *obj_args: Any
) -> Array:
    converted_fn, aux_pars = jax.closure_convert(objective_fn, init_pars, *obj_args)
    # aux_pars seems to be empty? took that line from jax docs example...
    solver = jaxopt.OptaxSolver(
        fun=converted_fn, opt=optax.adam(lr), implicit_diff=True, maxiter=5000
    )
    return solver.run(init_pars, *obj_args, *aux_pars)[0]


def global_fit_objective(data: Array, model: pyhf.Model) -> Callable[[Array], float]:
    def fit_objective(lhood_pars_to_optimize: Array) -> float:  # NLL
        """lhood_pars_to_optimize: either all pars, or just nuisance pars"""
        return cast(
            float, -model.logpdf(lhood_pars_to_optimize, data)[0]
        )  # pyhf.Model.logpdf returns list[float]

    return fit_objective


@partial(jax.jit, static_argnames=["model"])
def fit(
    data: Array,
    model: pyhf.Model,
    init_pars: Array,
    lr: float = 1e-3,
) -> Array:
    obj = global_fit_objective(data, model)
    fit_res = _minimize(obj, init_pars, lr)
    return fit_res


def fixed_poi_fit_objective(
    data: Array,
    model: pyhf.Model,
) -> Callable[[Array, float], float]:
    poi_idx = model.config.poi_index

    def fit_objective(
        lhood_pars_to_optimize: Array, poi_condition: float
    ) -> float:  # NLL
        """lhood_pars_to_optimize: either all pars, or just nuisance pars"""
        # pyhf.Model.logpdf returns list[float]
        blank = jnp.zeros_like(jnp.asarray(model.config.suggested_init()))
        blank += lhood_pars_to_optimize
        return cast(float, -model.logpdf(blank.at[poi_idx].set(poi_condition), data)[0])

    return fit_objective


@partial(jax.jit, static_argnames=["model"])
def fixed_poi_fit(
    data: Array,
    model: pyhf.Model,
    init_pars: Array,
    poi_condition: float,
    lr: float = 1e-3,
) -> Array:
    obj = fixed_poi_fit_objective(data, model)
    fit_res = _minimize(obj, init_pars, lr, poi_condition)
    blank = jnp.zeros_like(jnp.asarray(model.config.suggested_init()))
    blank += fit_res
    poi_idx = model.config.poi_index
    return blank.at[poi_idx].set(poi_condition)



from relaxed._version import version as __version__

__all__ = (
    "__version__",
    "hist",
    "cramer_rao_uncert",
    "fisher_info",
    "mle",
    "infer",
    "metrics",
    "cut",
    "dummy_pyhf",
)

from relaxed import dummy_pyhf, infer, metrics, mle
from relaxed.ops import cramer_rao_uncert, cut, fisher_info, hist



import jax.numpy as jnp
import numpy as np
import pyhf
import pytest
from jax import jacrev
from jax.random import PRNGKey

import relaxed
from relaxed.dummy_pyhf import example_model


@pytest.fixture
def data():
    nBg = 8000
    nSig = 300
    bins = np.linspace(0, 70, 10)
    background = np.histogram(np.random.normal(40, 10, nBg), bins=bins)[0]
    signal = np.histogram(np.random.normal(50, 5, nSig), bins=bins)[0]
    return signal, background


def test_gaussianity():
    pyhf.set_backend("jax")
    m = pyhf.simplemodels.uncorrelated_background([5, 5], [50, 50], [5, 5])
    pars = jnp.asarray(m.config.suggested_init())
    data = jnp.asarray(m.expected_data(pars))
    relaxed.metrics.gaussianity(m, pars, data, PRNGKey(0))


def test_gaussianity_grad():
    def pipeline(x):
        model = example_model(5.0)
        pars = model.config.suggested_init()
        data = model.expected_data(pars)
        return relaxed.metrics.gaussianity(model, pars * x, data * x, PRNGKey(0))

    jacrev(pipeline)(4.0)  # just check you can calc it w/o exception


def test_significance(data):
    signal, background = data

    def one_bin_significance(S, B):
        return jnp.sqrt(2 * ((S + B) * jnp.log(1 + S / B) - S))

    # non-zero signal bin
    assert np.allclose(
        one_bin_significance(signal[6], background[6]),
        relaxed.metrics.asimov_sig(signal[6], background[6]),
    )

    relaxed.metrics.asimov_sig(signal, background)


def test_significance_grad(data):
    def pipeline(
        pars, bw, endpoints  # bin widths up to the last one (determined by endpoints)
    ):
        s, b = data
        start, end = endpoints
        bins = jnp.cumsum(jnp.array([start, *pars, (end - start) - jnp.sum(pars)]))
        sig_hist = relaxed.hist(s, bins=bins, bandwidth=bw)
        bg_hist = relaxed.hist(b, bins=bins, bandwidth=bw)
        sig = relaxed.metrics.asimov_sig(sig_hist, bg_hist)
        return sig

    jacrev(pipeline)(
        jnp.array([5.0, 10.0, 15.0]), 1e-3, (0, 70)
    )  # just check you can calc it w/o exception



from __future__ import annotations

__all__ = ("example_model", "uncorrelated_background")

from typing import Any, Iterable

import jax.numpy as jnp
import pyhf


# class-based
class _Config:
    def __init__(self) -> None:
        self.poi_index = 0
        self.npars = 2

    def suggested_init(self) -> jnp.ndarray:
        return jnp.asarray([1.0, 1.0])


class Model:
    """Dummy class to mimic the functionality of `pyhf.Model`."""

    def __init__(self, spec: Iterable[Any]) -> None:
        pyhf.set_backend("jax")
        self.sig, self.nominal, self.uncert = spec
        self.factor = (self.nominal / self.uncert) ** 2
        self.aux = 1.0 * self.factor
        self.config = _Config()

    def expected_data(self, pars: jnp.ndarray) -> jnp.ndarray:
        mu, gamma = pars
        expected_main = jnp.asarray([gamma * self.nominal + mu * self.sig])
        aux_data = jnp.asarray([self.aux])
        return jnp.concatenate([expected_main, aux_data])

    def logpdf(self, pars: jnp.ndarray, data: jnp.ndarray) -> jnp.ndarray:
        maindata, auxdata = data
        main, _ = self.expected_data(pars)
        _, gamma = pars
        main = pyhf.probability.Poisson(main).log_prob(maindata)
        constraint = pyhf.probability.Poisson(gamma * self.factor).log_prob(auxdata)
        # sum log probs over bins
        return [jnp.sum(jnp.asarray([main + constraint]), axis=None)]


def uncorrelated_background(s: jnp.ndarray, b: jnp.ndarray, db: jnp.ndarray) -> Model:
    """Dummy class to mimic the functionality of `pyhf.simplemodels.hepdata_like`."""
    return Model([s, b, db])


def _calc_yields(x: jnp.ndarray) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    s = 15 + x
    b = 45 - 2 * x
    db = 1 + 0.2 * x**2
    return [s], [b], [db]


def example_model(
    phi: jnp.ndarray, return_yields: bool = False
) -> Model | tuple[Model, jnp.ndarray]:
    s, b, db = yields = _calc_yields(phi)

    model = uncorrelated_background(
        jnp.asarray([s]), jnp.asarray([b]), jnp.asarray([db])
    )

    if return_yields:
        return model, yields
    else:
        return model



import jax.numpy as jnp
import numpy as np
import pyhf
import pytest
from jax import jacrev

import relaxed
from relaxed.dummy_pyhf import example_model, uncorrelated_background


@pytest.mark.parametrize("test_stat", ["q", "q0"])
@pytest.mark.parametrize("phi", np.linspace(0.0, 10.0, 5))
def test_hypotest_validity(phi, test_stat):
    pyhf.set_backend("jax")
    if test_stat == "q":
        analytic_pars = jnp.array([0.0, 1.0])  # bkg-only hypothesis
    elif test_stat == "q0":
        analytic_pars = jnp.array([1.0, 1.0])  # nominal sig+bkg hypothesis
    else:
        raise ValueError(f"Unknown test statistic: {test_stat}")
    model, yields = example_model(phi, return_yields=True)
    relaxed_cls = relaxed.infer.hypotest(
        1, model.expected_data(analytic_pars), model, lr=1e-3, test_stat=test_stat
    )
    m = pyhf.simplemodels.uncorrelated_background(*yields)
    pyhf_cls = pyhf.infer.hypotest(
        1, m.expected_data(analytic_pars), m, test_stat=test_stat
    )
    assert np.allclose(
        relaxed_cls,
        pyhf_cls,
        atol=0.001,
    )  # tested working without dummy_pyhf on a pyhf fork, but not main yet


@pytest.mark.parametrize("test_stat", ["q", "q0"])
def test_hypotest_expected(test_stat):
    pyhf.set_backend("jax")
    if test_stat == "q":
        analytic_pars = jnp.array([0.0, 1.0])  # bkg-only hypothesis
    elif test_stat == "q0":
        analytic_pars = jnp.array([1.0, 1.0])  # nominal sig+bkg hypothesis
    else:
        raise ValueError(f"Unknown test statistic: {test_stat}")
    model, yields = example_model(5.0, return_yields=True)
    relaxed_cls = relaxed.infer.hypotest(
        1,
        model.expected_data(analytic_pars),
        model,
        lr=1e-3,
        test_stat=test_stat,
        expected_pars=analytic_pars,
    )
    m = pyhf.simplemodels.uncorrelated_background(*yields)
    pyhf_cls = pyhf.infer.hypotest(
        1, m.expected_data(analytic_pars), m, test_stat=test_stat
    )
    assert np.allclose(
        relaxed_cls,
        pyhf_cls,
        atol=0.001,
    )  # tested working without dummy_pyhf on a pyhf fork, but not main yet


@pytest.mark.parametrize("test_stat", ["q", "q0"])
@pytest.mark.parametrize("expected_pars", [True, False])
def test_hypotest_grad(test_stat, expected_pars):
    pars = jnp.array([0.0, 1.0])
    if expected_pars:
        expars = pars
    else:
        expars = None

    def pipeline(x):
        model = uncorrelated_background(x * 5.0, x * 20, x * 2)
        expected_cls = relaxed.infer.hypotest(
            1.0,
            model=model,
            data=model.expected_data(pars),
            lr=1e-2,
            test_stat=test_stat,
            expected_pars=expars,
        )
        return expected_cls

    jacrev(pipeline)(jnp.asarray(0.5))


def test_wrong_test_stat():
    with pytest.raises(ValueError):
        model = example_model(0.0)
        relaxed.infer.hypotest(
            1,
            model.expected_data(jnp.array([0.0, 1.0])),
            model,
            lr=1e-2,
            test_stat="q1",
        )



from __future__ import annotations

__all__ = ("cut", "hist", "fisher_info", "cramer_rao_uncert")

from functools import partial
from typing import Any

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from relaxed._types import Array


@partial(jax.jit, static_argnames=["keep"])
def cut(data: Array, cut_val: float, slope: float = 1.0, keep: str = "above") -> Array:
    """Use a sigmoid function as an approximate cut. Same as a hard cut in the limit of infinite slope.
    Note: this function returns weights, not indices.

    Parameters
    ----------
    data : Array
        The data to cut.
    cut_val : float
        The value to cut at.
    slope : float
        The slope of the sigmoid function.
    keep : str, optional
        Whether to keep the data above or below the cut. One of:
        - "above" (default)
        - "below"

    Returns
    -------
    Array
        Weighted yields of the data after the cut.
    """
    if keep == "above":
        return 1 / (1 + jnp.exp(-slope * (data - cut_val)))
    elif keep == "below":
        return 1 / (1 + jnp.exp(slope * (data - cut_val)))
    else:
        raise ValueError(f"keep must be one of 'above' or 'below', not {keep}")


@partial(jax.jit, static_argnames=["density", "reflect_infinities"])
def hist(
    data: Array,
    bins: Array,
    bandwidth: float,  # | None = None,
    density: bool = False,
    reflect_infinities: bool = False,
) -> Array:
    """Differentiable histogram, defined via a binned kernel density estimate (bKDE).

    Parameters
    ----------
    data : Array
        1D array of data to histogram.
    bins : Array
        1D array of bin edges.
    bandwidth : float
        The bandwidth of the kernel. Bigger == lower gradient variance, but more bias.
    density : bool
        Normalise the histogram to unit area.
    reflect_infinities : bool
        If True, define bins at +/- infinity, and reflect their mass into the edge bins.

    Returns
    -------
    Array
        1D array of bKDE counts.
    """
    # bandwidth = bandwidth or events.shape[-1] ** -0.25  # Scott's rule

    bins = jnp.array([-jnp.inf, *bins, jnp.inf]) if reflect_infinities else bins

    # get cumulative counts (area under kde) for each set of bin edges
    cdf = jsp.stats.norm.cdf(bins.reshape(-1, 1), loc=data, scale=bandwidth)
    # sum kde contributions in each bin
    counts = (cdf[1:, :] - cdf[:-1, :]).sum(axis=1)

    if density:  # normalize by bin width and counts for total area = 1
        db = jnp.array(jnp.diff(bins), float)  # bin spacing
        counts = counts / db / counts.sum(axis=0)

    if reflect_infinities:
        counts = (
            counts[1:-1]
            + jnp.array([counts[0]] + [0] * (len(counts) - 3))
            + jnp.array([0] * (len(counts) - 3) + [counts[-1]])
        )

    return counts


@partial(jax.jit, static_argnames=["model"])
def fisher_info(model: Any, pars: Array, data: Array) -> Array:
    """Fisher information matrix for a model with a logpdf method.

    Parameters
    ----------
    model : Any
        The model to compute the Fisher information matrix for.
        Needs to have a logpdf method (that returns list[float] for now).
    pars : Array
        The (MLE) parameters of the model.
    data : Array
        The data to compute the Fisher information matrix for.

    Returns
    -------
    Array
        Fisher information matrix.
    """
    return jnp.linalg.inv(-jax.hessian(lambda p, d: model.logpdf(p, d)[0])(pars, data))


@partial(jax.jit, static_argnames=["model"])
def cramer_rao_uncert(model: Any, pars: Array, data: Array) -> Array:
    """Approximate uncertainties on MLE parameters for a model with a logpdf method.
    Defined as the square root of the diagonal of the Fisher information matrix, valid
    via the Cramer-Rao lower bound.

    Parameters
    ----------
    model : Any
        The model to compute the Cramer-Rao uncertainty for.
        Needs to have a logpdf method (that returns list[float] for now).
    pars : Array
        The (MLE) parameters of the model.
    data : Array
        The data to compute the uncertainty for.

    Returns
    -------
    Array
        Cramer-Rao uncertainty on the MLE parameters.
    """
    return jnp.sqrt(jnp.diag(fisher_info(model, pars, data)))



"""Calculate expected CLs values with hypothesis tests."""
from __future__ import annotations

__all__ = ("hypotest",)

import logging
from functools import partial

import jax.numpy as jnp
import pyhf
from jax import jit

from relaxed._types import Array
from relaxed.mle import fit, fixed_poi_fit


def hypotest(
    test_poi: float,
    data: Array,
    model: pyhf.Model,
    lr: float,
    return_mle_pars: bool = False,
    test_stat: str = "q",
    expected_pars: Array | None = None,
) -> tuple[Array, Array] | Array:
    """Calculate expected CLs/p-values via hypothesis tests.

    Parameters
    ----------
    test_poi : float
        The value of the test parameter to use for the hypothesis test.
    data : Array
        The data to use for the hypothesis test.
    model : pyhf.Model
        The model to use for the hypothesis test.
    lr : float
        Learning rate for the MLE fit, done via gradient descent.
    return_mle_pars : bool, optional
        Whether to return the MLE parameters calculated as a by-product.
    test_stat : str, optional
        The test statistic to use for the hypothesis test. One of:
        - "qmu" (default, used for upper limits)
        - "q0" (used for discovery of a positive signal)
    expected_pars : Array, optional
        Use if calculating expected significance and these are known. If not
        provided, the MLE parameters will be fitted.

    Returns
    -------
    Array
        The expected CLs/p-value.
    Array
        The MLE parameters, if `return_mle_pars` is True.
    """
    if test_stat == "q":
        return qmu_test(test_poi, data, model, lr, return_mle_pars, expected_pars)
    elif test_stat == "q0":
        logging.info(
            "test_poi automatically set to 0 for q0 test (bkg-only null hypothesis)"
        )
        return q0_test(0.0, data, model, lr, return_mle_pars, expected_pars)
    else:
        raise ValueError(f"Unknown test statistic: {test_stat}")


@partial(
    jit, static_argnames=["model", "return_mle_pars"]
)  # can remove model eventually
def qmu_test(
    test_poi: float,
    data: Array,
    model: pyhf.Model,
    lr: float,
    return_mle_pars: bool = False,
    expected_pars: Array | None = None,
) -> tuple[Array, Array] | Array:
    # hard-code 1 as inits for now
    # TODO: need to parse different inits for constrained and global fits
    # because init_pars[0] is not necessarily the poi init
    init_pars = jnp.asarray(model.config.suggested_init())
    conditional_pars = fixed_poi_fit(
        data, model, poi_condition=test_poi, init_pars=init_pars[:-1], lr=lr
    )
    if expected_pars is None:
        mle_pars = fit(data, model, init_pars=init_pars, lr=lr)
    else:
        mle_pars = expected_pars
    profile_likelihood = -2 * (
        model.logpdf(conditional_pars, data)[0] - model.logpdf(mle_pars, data)[0]
    )

    poi_hat = mle_pars[model.config.poi_index]
    qmu = jnp.where(poi_hat < test_poi, profile_likelihood, 0.0)

    CLsb = 1 - pyhf.tensorlib.normal_cdf(jnp.sqrt(qmu))
    altval = 0.0
    CLb = 1 - pyhf.tensorlib.normal_cdf(altval)
    CLs = CLsb / CLb
    return (CLs, mle_pars) if return_mle_pars else CLs


@partial(
    jit, static_argnames=["model", "return_mle_pars"]
)  # can remove model eventually
def q0_test(
    test_poi: float,
    data: Array,
    model: pyhf.Model,
    lr: float,
    return_mle_pars: bool = False,
    expected_pars: Array | None = None,
) -> tuple[Array, Array] | Array:
    # hard-code 1 as inits for now
    # TODO: need to parse different inits for constrained and global fits
    # because init_pars[0] is not necessarily the poi init
    init_pars = jnp.asarray(model.config.suggested_init())
    conditional_pars = fixed_poi_fit(
        data, model, poi_condition=test_poi, init_pars=init_pars[:-1], lr=lr
    )
    if expected_pars is None:
        mle_pars = fit(data, model, init_pars=init_pars, lr=lr)
    else:
        mle_pars = expected_pars
    profile_likelihood = -2 * (
        model.logpdf(conditional_pars, data)[0] - model.logpdf(mle_pars, data)[0]
    )

    poi_hat = mle_pars[model.config.poi_index]
    q0 = jnp.where(poi_hat >= test_poi, profile_likelihood, 0.0)
    p0 = 1 - pyhf.tensorlib.normal_cdf(jnp.sqrt(q0))

    return (p0, mle_pars) if return_mle_pars else p0
